import { Injectable, EventEmitter } from '@angular/core';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { Observable, Subject, BehaviorSubject, of, from, interval } from 'rxjs';
import { fromEvent, throwError, Subscription, merge } from 'rxjs';
import { catchError, map, tap, filter } from 'rxjs/operators';
import { Job } from './job';
import { AppAction, Strudelapp, StrudelappInstance } from './strudelapp';
import { Computesite, Health } from './computesite';
import { Identity, AuthToken, KeyCert, SshAuthzServer } from './identity';
import { BatchInterface } from './batchinterface';
import { repeat, takeUntil, take, switchMap } from 'rxjs/operators';
import { LocationStrategy, Location } from '@angular/common';
import { ActivatedRoute, Router } from '@angular/router';
import { AuthorisationService } from './authorisation.service';
import { BackendSelectionService } from './backend-selection.service';
import {BrowserWindowService} from './browser-window.service';
import { NotificationsService } from './notifications.service';
import { APIServer } from './apiserver';

/** The TES service contains ways to start Tunnels, and Execute programs
Its also responsible for querying a compute site for running jobs */



@Injectable({
  providedIn: 'root',
})
export class TesService {
public Base: string;
private twsproxy: string;
public jobs: any[];
private appwindow: any;
private updateJobSub: Subscription;
private updateUserHealthSub: Subscription;
private cachetincidents: BehaviorSubject<Health[]>;
private cancelRequests$: Subject<boolean>;
public openWindow$: Subject<any>;


  constructor(private http: HttpClient,
              private authorisationService: AuthorisationService,
              private backendSelectionService: BackendSelectionService,
              private location: Location,
              private browserWindow: BrowserWindowService,
              private notifications: NotificationsService,
  ) {
    this.cancelRequests$ = new Subject<boolean>();
    this.openWindow$ = new Subject<any>();
    this.backendSelectionService.apiserver.subscribe( (value) => { if (value != null) {this.twsproxy = value.tws ; this.Base = value.tes }});
 }



 private buildParams(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, appinst?: any): string {
   let params = new URLSearchParams();
   let id = identity.copy_skip_catalog();
   id.site.appCatalog = null;
   if (appinst!== null)  {
     params.set('appinstance',JSON.stringify(appinst));
   }
   params.set('app',JSON.stringify(app));
   params.set('interface',JSON.stringify(batchinterface));
   params.set('identity',JSON.stringify(id));
   return params.toString();
 }


 public cancelHealthRequests() {
    this.cancelRequests$.next(true);
 }

 private getUserHealthError(error: any, identity: Identity) {
   identity.accountalerts.next([]);
   this.notifications.notify("There was an error checking your user account");
 }


getUserHealth(identity: Identity)  {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams();
  if (identity === undefined || identity === null) {
    return
  }
  if (identity.site.userhealth === undefined) {
    identity.accountalerts.next([]);
    return
  }
  params.set('statcmd',JSON.stringify(identity.site.userhealth));
  params.set('host',JSON.stringify(identity.site.host));
  params.set('username',JSON.stringify(identity.username));
  
  this.updateUserHealthSub = this.http.get<Health[]>(this.Base+'/stat'+'?'+params.toString(),options)
    .pipe(takeUntil(this.cancelRequests$))
    .subscribe(resp => this.addUserHealth(identity,resp), error => this.getUserHealthError(error,identity));
}

public getHealthAlerts(identity: Identity) {
  this.getCachetIncidents(identity);
  this.getUserHealth(identity);
}

public getCachetIncidents(identity: Identity) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: false};
  let params = new URLSearchParams();

  if (identity.site.cacheturis === undefined ||  identity.site.cacheturis.length == 0) {
    identity.systemalerts.next([]);
    return
  }
  for (let uri of identity.site.cacheturis) {
    this.http.get(uri,options)
      .pipe(takeUntil(this.cancelRequests$))
      .subscribe(resp => this.addCachetIncidents(identity,resp), error => this.getCachetIncidentsError(error,identity));
  }
}

private addCachetIncidents(identity,resp) {
  let ci = [];
  for (let i of resp.data) {
    if (i.status == 3 || i.status == 4) {
      continue;
    }
    let h = new Health();
    h.stat = 'error';
    h.msg = i.message;
    ci.push(h);
  }
  identity.systemalerts.next(ci);
}

private addUserHealth(identity,resp) {
  let ci = []
  for (let i of resp) {
    let h = new Health();
    h.stat = i.stat;
    h.msg = i.message;
    if (i.title != undefined) {
      h.title = i.title;
    }
    if (i.type != undefined) {
      h.type = i.type;
      h.data = i.data;
    }
    ci.push(h);
  }
  identity.accountalerts.next(ci);
}

 private getCachetIncidentsError(error: any, identity: Identity) {
   if (error.status == 0) {
     return
   }
   console.error(error);
   if (error.status == 401) {
     this.notifications.notify("Login expired. Please log in again.");
     this.authorisationService.updateAgentContents();
     return
   }
   if (error.status == 400) {
       if (error.error !== undefined && error.error.message !== undefined) {
        this.notifications.notify(error.error.message);
       }
       return
   }
 }


 submissionError(error: any) {
   if (error.status != 0) {
       if ('error' in error && 'message' in error.error) {
         this.notifications.notify(error.error.message);
       } else {
         this.notifications.notify('Job submission failed');
       }
   }
 }

 cancelError(error: any) {
   if (error.status != 0) {
       if ('error' in error && 'message' in error.error) {
         this.notifications.notify(error.error.message);
       } else {
         this.notifications.notify('Job Canceling failed');
       }
   }
 }

 buildBody(app: Strudelapp, appparams?: string) {
   return JSON.stringify({'app': app, 'appparams': appparams});
 }

 submit(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, ftidentities?: Identity[], appparams?: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   this.notifications.notify('Submitting job');
   let paramstr = this.buildParams(app,identity,batchinterface);
   let keys = JSON.stringify(this.authorisationService.getKeys());
   let loggedin = JSON.stringify(this.authorisationService.loggedInAuthZ.value);
   let ids = [];
   if (ftidentities != undefined) {
     for (let id of ftidentities) {
         ids.push(id.copy_skip_catalog())
     }
   }
   let body = {'app': app, 'appparams': appparams, 'keys': keys, 'ids': JSON.stringify(JSON.stringify(ids))}
   this.http.post<any>(this.Base+'/submit'+'?'+paramstr, body, options)
                                          .subscribe(resp => { this.notifications.notify(null) },
                                                     error => this.submissionError(error));
 }


 cancel(job: Job) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   let data = {};
   let bi = new BatchInterface();
   bi.statcmd = job.identity.site.statcmd;
   bi.cancelcmd = job.identity.site.cancelcmd;
   let paramstr = this.buildParams(job.app,job.identity,bi);
   this.http.delete<any>(this.Base+'/cancel/'+job.jobid+'?'+paramstr, options)
     .subscribe(resp => {} ,
                    error => this.cancelError(error));
 }

 public getAppInstance(job: Job, action: AppAction, apiserver: APIServer) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let jobid = job.jobid;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   params.set('cmd',JSON.stringify(action.paramscmd));
   let paramstr = params.toString();
   return this.http.get<string>(apiserver.tes+'/appinstance/'+username+'/'+loginhost+'/'+batchhost+'/'+jobid+'?'+paramstr, options)
 }

 public createTunnel(job: Job, appinst: any, action: AppAction, apiserver: APIServer) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   job.connectionState = 2;

   if ((action.notunnel != undefined) && (action.notunnel == true)) {
     return of([])
   }
   let options = { headers: headers, withCredentials: true};
   let paramstr = params.toString();
   return this.http.post<string>(apiserver.tes+'/createtunnel/'+username+'/'+loginhost+'/'+batchhost+'?'+paramstr, appinst, options)
 }

 public getAppUrl(job: Job, appinst: any, action: AppAction, apiserver: APIServer) {
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   if (action.client === undefined || action.client === null) {
     return of(null)
   }
   let pseudoapp = {'client':{'redir':action.client.redir,'cmd':action.client.cmd}};
   params.set('app',JSON.stringify(pseudoapp));
   params.set('appinst',JSON.stringify(appinst));
   let paramstr = params.toString();
   job.connectionState = 3;
   if (action.client.cmd !== undefined && action.client.cmd !== null) {
      return this.http.get<string>(apiserver.tes+'/applaunch?'+paramstr,options)
   }
   return this.http.get<string>(apiserver.tes+'/appurl?'+paramstr,options)
 }

 public contactUs(form: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   return this.backendSelectionService.apiserver.pipe(
     filter((v) => v !== undefined && v !== null),
     switchMap((apiserver) => this.http.post(apiserver.tes+'/contactus',{'data':form},options))
   )
 }

 public connect(job: Job, action: AppAction, appinst?: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};

   job.connectionState=1
   // Can no connect till we know which api server we are using. Threfore the apiserver
   // is the start of the pipe
   // the API server is behaviorsubject so it should always fire immediately
   this.backendSelectionService.apiserver.pipe(
     filter((v) => v !== undefined && v !== null),
     switchMap((apiserver) => this.getAppInstance(job,action,apiserver),
       (apiserver,appinst,i1,i2) => {return [apiserver,appinst]}),
     tap((_) => job.connectionState=2),
     switchMap(([apiserver,appinst]) => { return this.createTunnel(job,appinst,action,<APIServer>apiserver)},
       (data,tunnelresp) => { data[1]['localport'] = (<any> tunnelresp)['localport'] ; return data}), // we don't care about data from the tunnel as long as it was successful.`
     tap((_) => job.connectionState=3),
     switchMap(([apiserver,appinst]) => this.getAppUrl(job,appinst,action,<APIServer>apiserver),
       ([apiserver,appinst],url) => {return [apiserver,appinst,url]}),
     tap((_) => job.connectionState=4),
   ).subscribe(([apiserver,appinst,url]) => { if (url !== null) {
                                          this.openWindow$.next( {'job':job, 
                                                'url': <string> url, 
                                                'usebasicauth':action.client.usebasicauth,
                                                'apiserver':apiserver,
                                            'action':action,'appinst':appinst});
                                      }
                                    job.connectionState=0},
     (err) => { job.connectionState = 0; this.handleError(err)})
 }


  public runCommand(identity: Identity, command: string): Observable<any> {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  // remove from the job list any jobs for identities that we don't know about
  let params = new URLSearchParams();
  if (identity === undefined || identity === null) {
    return
  }
  params.set('statcmd',JSON.stringify(command));
  params.set('host',JSON.stringify(identity.site.host));
  params.set('username',JSON.stringify(identity.username));
  
  return  this.http.get(this.Base+'/stat'+'?'+params.toString(),options)

}


public postMkDir(id: Identity, path: string, name: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));

  return this.http.post<any>(this.Base+'/mkdir?'+params.toString(),name, options);
}

public getSftpData(id: Identity, path: string, cd: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));
  params.set('cd',JSON.stringify(cd));

  return this.http.get<any>(this.Base+'/ls?'+params.toString(),options)
                      // .pipe(catchError(this.handleError))

}

private handleError(error: HttpErrorResponse) {
  console.error(error);
  if (error.error instanceof ErrorEvent) {
    this.notifications.notify("A networking error occured.")
    return
  } else {
    var re = /identity/gi;
    if (error.error.message != undefined) {
      let searchresult = error.error.message.search(re);
      if (searchresult != -1) {
        this.notifications.notify('Login expired, refreshing');
        this.authorisationService.updateAgentContents();
        return
      }
      return this.notifications.notify(error.error.message)
    } 
  }
}





}
